# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Notifications is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""

import pytest
from invenio_app.factory import create_app as _create_app


@pytest.fixture(scope="module")
def app_config(app_config):
    """Override pytest-invenio app_config fixture."""
    # TODO: Override any necessary config values for tests
    # app_config["NOTIFICATIONS_FOOBAR"] = "test"
    return app_config


@pytest.fixture(scope="module")
def create_app(instance_path):
    """Application factory fixture."""
    return _create_app



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generators for notification context."""

from abc import ABC, abstractmethod

from invenio_records.dictutils import dict_lookup, dict_set

from invenio_notifications.backends.email import EmailNotificationBackend
from invenio_notifications.registry import EntityResolverRegistry


class ContextGenerator(ABC):
    """Payload generator for a notification."""

    @abstractmethod
    def __call__(self, notification):
        """Update notification context."""
        raise NotImplementedError()


class RecipientGenerator(ABC):
    """Recipient generator for a notification."""

    @abstractmethod
    def __call__(self, notification, recipients):
        """Add recipients."""
        raise NotImplementedError()


class RecipientBackendGenerator(ABC):
    """Backend generator for a notification."""

    @abstractmethod
    def __call__(self, notification, recipient, backends):
        """Update required recipient information and add backend id."""
        raise NotImplementedError()


class EntityResolve(ContextGenerator):
    """Payload generator for a notification using the entity resolvers."""

    def __init__(self, key):
        """Ctor."""
        self.key = key

    def __call__(self, notification):
        """Update required recipient information and add backend id."""
        entity_ref = dict_lookup(notification.context, self.key)
        entity = EntityResolverRegistry.resolve_entity(entity_ref)
        dict_set(notification.context, self.key, entity)
        return notification


class UserEmailBackend(RecipientBackendGenerator):
    """User related email backend generator for a notification."""

    def __call__(self, notification, recipient, backends):
        """Add backend id to backends."""
        backend_id = EmailNotificationBackend.id
        backends.append(backend_id)
        return backend_id



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""E-mail specific notification backend."""

from flask import current_app
from invenio_mail.tasks import send_email

from invenio_notifications.backends.base import NotificationBackend
from invenio_notifications.backends.utils.loaders import JinjaTemplateLoaderMixin


class EmailNotificationBackend(NotificationBackend, JinjaTemplateLoaderMixin):
    """Email specific notification backend."""

    id = "email"

    def send(self, notification, recipient):
        """Mail sending implementation."""
        content = self.render_template(notification, recipient)

        resp = send_email(
            {
                "subject": content["subject"],
                "html": content["html_body"],
                "body": content["plain_body"],
                "recipients": [
                    recipient.data.get("email") or recipient.data.get("email_hidden")
                ],
                "sender": current_app.config["MAIL_DEFAULT_SENDER"],
            }
        )
        return resp  # TODO: what would a "delivery" result be



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notification manager."""


from invenio_notifications.tasks import broadcast_notification, dispatch_notification


class NotificationManager:
    """Notification manager.

    Taking care of building notifications and forwarding them to the backend(s).
    """

    def __init__(self, backends, builders):
        """Ctor."""
        self.backends = backends  # via config "NOTIFICATIONS_BACKENDS"
        self.builders = builders  # via config "NOTIFICATIONS_BUILDERS"

    # def validate(self, notification):
    #     """Validate notification object."""
    #     # Validate the type
    #     ...
    #     # Validate context (if possible)
    #     ...

    # Client
    def broadcast(self, notification, eager=False):
        """Broadcast a notification via a Celery task."""
        self.validate(notification)
        task = broadcast_notification.si(notification.dumps())
        return task.apply() if eager else task.delay()

    # Consumer
    def handle_broadcast(self, notification):
        """Handle a notification broadcast."""
        builder = self.builders[notification.type]
        # Resolve and expand entities
        builder.resolve_context(notification)
        # Generate recipients
        recipients = builder.build_recipients(notification)
        recipients = builder.filter_recipients(notification, recipients)
        for recipient in recipients.values():
            recipient_backends = builder.build_recipient_backends(
                notification, recipient
            )
            for backend in recipient_backends:
                dispatch_notification.delay(
                    backend,
                    recipient.dumps(),
                    notification.dumps(),
                )

    # Dispatch delivery/sending via a backend
    def handle_dispatch(self, backend_id, recipient, notification):
        """Handle a backend dispatch."""
        self.backends[backend_id].send(notification, recipient)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tasks related to notifications."""

from celery import shared_task

from invenio_notifications.models import Notification, Recipient

from .proxies import current_notifications_manager


@shared_task
def broadcast_notification(notification):
    """Handles a notification broadcast."""
    current_notifications_manager.handle_broadcast(Notification(**notification))


@shared_task(max_retries=5, default_retry_delay=5 * 60)
def dispatch_notification(backend, recipient, notification):
    """Dispatches a notification to a recipient for a specific backend."""
    current_notifications_manager.handle_dispatch(
        backend, Recipient(**recipient), Notification(**notification)
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Filters for notification recipients."""

from abc import ABC, abstractmethod


class RecipientFilter(ABC):
    """Recipient filter for a notification."""

    @abstractmethod
    def __call__(self, notification, recipients):
        """Filter recipients."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Notifications is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module for notifications support."""

from setuptools import setup

setup()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module for notifications support."""

from importlib_metadata import entry_points

from . import config
from .manager import NotificationManager


class InvenioNotifications(object):
    """Invenio-Notifications extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.init_manager(app)
        self.init_registries(app)
        app.extensions["invenio-notifications"] = self

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("NOTIFICATIONS_"):
                app.config.setdefault(k, getattr(config, k))

    def init_manager(self, app):
        """Initialize manager."""
        manager = NotificationManager(
            backends=app.config["NOTIFICATIONS_BACKENDS"],
            builders=app.config["NOTIFICATIONS_BUILDERS"],
        )
        self.manager = manager

    def init_registries(self, app):
        """Initialize registries."""
        self.entity_resolvers = {
            er.type_key: er for er in app.config["NOTIFICATIONS_ENTITY_RESOLVERS"]
        }
        for ep in entry_points(group="invenio_notifications.entity_resolvers"):
            er_cls = ep.load()
            self.entity_resolvers.setdefault(er_cls.type_key, er_cls())



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Notifications is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notifications views module."""


from flask import Blueprint
from flask_breadcrumbs import register_breadcrumb
from flask_login import login_required
from flask_menu import register_menu
from invenio_i18n import LazyString
from invenio_i18n import lazy_gettext as _
from invenio_theme.proxies import current_theme_icons


def create_blueprint(app):
    """Register blueprint routes on app."""
    blueprint = Blueprint(
        "invenio_notifications",
        __name__,
        template_folder="templates",
        static_folder="static",
    )

    return blueprint


def create_blueprint_settings(app):
    """Register blueprint routes on app."""
    blueprint = Blueprint(
        "invenio_notifications_settings",
        __name__,
        template_folder="templates/settings",
        url_prefix="/account/settings/notifications",
        static_folder="static",
    )

    notifications_settings_view_func = app.config[
        "NOTIFICATIONS_SETTINGS_VIEW_FUNCTION"
    ]

    if notifications_settings_view_func:

        @blueprint.route("/", methods=["GET", "POST"])
        @login_required
        @register_menu(
            blueprint,
            "settings.notifications",
            # NOTE: Menu item text (icon replaced by a bell icon).
            _(
                "%(icon)s Notifications",
                icon=LazyString(lambda: f'<i class="{current_theme_icons.bell}"></i>'),
            ),
            order=2,
        )
        @register_breadcrumb(
            blueprint, "breadcrumbs.settings.notifications", _("Notifications")
        )
        def index():
            """View for notification settings."""
            return notifications_settings_view_func()

    return blueprint
